'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.isValidBankVerificationCode = exports.isValidBankAccountNumber = exports.isValidBankAgencyNumber = exports.isValidBankId = exports.isValidCNPJ = exports.isValidPhone = exports.isValidEmail = exports.isValidCPF = exports.isNotEmpty = exports.isPositiveNumber = exports.areNotEmpty = exports.replaceNonDigits = undefined;

var _validate = require('./validate.js');

var _isValidCPF = require('./isValidCPF.js');

var _isValidCNPJ = require('./isValidCNPJ.js');

var _isValidEmail = require('./isValidEmail.js');

var _isValidPhone = require('./isValidPhone.js');

var _errorMessages = require('./errorMessages.js');

var replaceNonDigits = exports.replaceNonDigits = function replaceNonDigits(value) {
  return String(value).replace(/\D/g, '');
};

var valueIsNotEmpty = function valueIsNotEmpty(value) {
  return value != null && value !== '';
};

var valueIsPositiveNumber = function valueIsPositiveNumber(value) {
  return !Number.isNaN(value) && Number(value) >= 0;
};

var valuesAreNotEmpty = function valuesAreNotEmpty(values) {
  return values.every(valueIsNotEmpty);
};

var valueIsValidEmail = function valueIsValidEmail(value) {
  return (0, _isValidEmail.isValidEmail)(String(value));
};

var valueIsValidPhone = function valueIsValidPhone(value) {
  return (0, _isValidPhone.isValidPhone)(String(value));
};

var valueIsValidCPF = function valueIsValidCPF(value) {
  return (0, _isValidCPF.isValidCPF)(replaceNonDigits(value));
};

var valueIsValidCNPJ = function valueIsValidCNPJ(value) {
  return (0, _isValidCNPJ.isValidCNPJ)(replaceNonDigits(value));
};

var valueIsNumeric = function valueIsNumeric(value) {
  return replaceNonDigits(String(value)) === String(value);
};

var valueLengthBetween = function valueLengthBetween(minLength, maxLength) {
  return function (value) {
    return String(value).length >= minLength && String(value).length <= maxLength;
  };
};

var valueGreaterThan = function valueGreaterThan(num) {
  return function (value) {
    return Number(value) > num;
  };
};

var areNotEmpty = exports.areNotEmpty = function areNotEmpty(scope) {
  return function (formdata) {
    return (0, _validate.validate)(scope, formdata).with(valuesAreNotEmpty, _errorMessages.MANDATORY_FIELDS_ALL);
  };
};

var isPositiveNumber = exports.isPositiveNumber = function isPositiveNumber(scope) {
  return function (formdata) {
    return (0, _validate.validate)(scope, formdata).with(valueIsPositiveNumber, _errorMessages.POSITIVE_NUMBER);
  };
};

var isNotEmpty = exports.isNotEmpty = function isNotEmpty(scope) {
  return function (formdata) {
    return (0, _validate.validate)(scope, formdata).with(valueIsNotEmpty, _errorMessages.MANDATORY_FIELD);
  };
};

var isValidCPF = exports.isValidCPF = function isValidCPF(scope) {
  return function (formdata) {
    return (0, _validate.validate)(scope, formdata).with(valueIsValidCPF, _errorMessages.INVALID_CPF);
  };
};

var isValidEmail = exports.isValidEmail = function isValidEmail(scope) {
  return function (formdata) {
    return (0, _validate.validate)(scope, formdata).with(valueIsValidEmail, _errorMessages.INVALID_EMAIL);
  };
};

var isValidPhone = exports.isValidPhone = function isValidPhone(scope) {
  return function (formdata) {
    return (0, _validate.validate)(scope, formdata).with(valueIsValidPhone, _errorMessages.INVALID_PHONE);
  };
};

var isValidCNPJ = exports.isValidCNPJ = function isValidCNPJ(scope) {
  return function (formdata) {
    return (0, _validate.validate)(scope, formdata).with(valueIsValidCNPJ, _errorMessages.INVALID_CNPJ);
  };
};

var isValidBankId = exports.isValidBankId = function isValidBankId(scope) {
  return function (formdata) {
    return (0, _validate.validate)(scope, formdata).with([valueIsNumeric, valueGreaterThan(0)], _errorMessages.MASK_ERROR);
  };
};

var isValidBankAgencyNumber = exports.isValidBankAgencyNumber = function isValidBankAgencyNumber(scope) {
  return function (formdata) {
    return (0, _validate.validate)(scope, formdata).with([valueIsNumeric, valueLengthBetween(1, 5), valueGreaterThan(0)], _errorMessages.MASK_ERROR);
  };
};

var isValidBankAccountNumber = exports.isValidBankAccountNumber = function isValidBankAccountNumber(scope) {
  return function (formdata) {
    return (0, _validate.validate)(scope, formdata).with([valueIsNumeric, valueLengthBetween(1, 12)], _errorMessages.MASK_ERROR);
  };
};

var isValidBankVerificationCode = exports.isValidBankVerificationCode = function isValidBankVerificationCode(scope) {
  return function (formdata) {
    return (0, _validate.validate)(scope, formdata).with(valueLengthBetween(1, 2), _errorMessages.MASK_ERROR);
  };
};